NAME = "Data"

ID = "orange.widgets.data"

DESCRIPTION = """Widgets for data manipulation."""

LONG_DESRIPTION = """
This category contains widgets for data manipulation. This includes
loading, importing, saving, preprocessing, selection, etc.

"""

ICON = "icons/Category-Data.svg"

BACKGROUND = "#FFD39F"

PRIORITY = 1
